import sys
import json

from policyuniverse.expander_minimizer import expand_policy
from policy_sentry.querying.actions import get_action_data
from policy_sentry.querying.actions import get_actions_matching_arn_type
from policy_sentry.querying.actions import get_actions_that_support_wildcard_arns_only

from aws_iam_utils.constants import READ, LIST, WRITE, WILDCARD_ARN_TYPE
from aws_iam_utils.util import extract_policy_permission_items

def policies_are_equal(p1, p2):
    """
    Checks whether two policies give the same permissions. This will expand all wildcards and Resource constraints and then compare the result.

    @param p1  The first policy. Should be a dict that contains a Statement key, which should be a list of dicts conforming to the AWS IAM Policy schema.
    @param p2  The second policy, same format as p1.

    @returns True if p1 and p2 represent exactly the same permissions, or False otherwise.
    """
    return extract_policy_permission_items(expand_policy(p1)) == extract_policy_permission_items(expand_policy(p2))

def policy_has_only_these_access_levels(p, access_levels):
    """
    Returns True if all actions granted under the given policy are Read or List actions.
    """
    p_items = extract_policy_permission_items(expand_policy(p))
    for item in p_items:
        action_service, action_name = item["action"].split(":")

        action_output = get_action_data(action_service, action_name)

        if action_output is False:
            raise ValueError(f'invalid action: {item["action"]}')

        for action_output_action in action_output[action_service]:
            if action_output_action["action"].lower() != item["action"].lower():
                continue

            if action_output_action["access_level"] not in access_levels:
                return False

    return True


def is_read_only_policy(p):
    """
    Returns True if all actions granted under the given policy are Read or List actions.
    """
    return policy_has_only_these_access_levels(p, [ READ, LIST ])


def is_list_only_policy(p):
    """
    Returns True if all actions granted under the given policy are List actions.
    """
    return policy_has_only_these_access_levels(p, [ LIST ])


def is_read_write_policy(p):
    """
    Returns True if all actions granted under the given policy are Read, List or Write actions.
    """
    return policy_has_only_these_access_levels(p, [ READ, LIST, WRITE ])

def policy_has_only_these_arn_types(p, service_name, arn_types):
    """
    Returns True if all actions granted under the given policy relate to the given ARN types only. Use `aws_iam_utils.constants.WILDCARD_ARN_TYPE` to refer to actions that do not relate to an ARN type (so-called "wildcard actions" in policy_sentry).
    """
    arn_type_actions = {}
    for arn_type in arn_types:
        if arn_type == WILDCARD_ARN_TYPE:
            arn_type_actions[arn_type] = [ x.lower() for x in get_actions_that_support_wildcard_arns_only(service_name) ]
        else:
            arn_type_actions[arn_type] = [ x.lower() for x in get_actions_matching_arn_type(service_name, arn_type) ]

    p_items = extract_policy_permission_items(expand_policy(p))
    for item in p_items:
        action_service, action_name = item["action"].split(":")

        action_output = get_action_data(action_service, action_name)

        if action_output is False:
            raise ValueError(f'invalid action: {item["action"]}')

        action_found = False
        for arn_type in arn_types:
            if item['action'].lower() in arn_type_actions[arn_type]:
                action_found = True
                break

        if action_found is False:
            return False

    return True





from aws_iam_utils.constants import READ, WRITE, LIST

"""
This is a mapping of actions to overridden action data. This is primarily used to provide missing/incorrect results from policy_sentry's get_action_data.
"""
ACTION_DATA_OVERRIDES = {
  # NB: keys must be all lowercase
  'lambda:getfunctionurlconfig': {
    'action': 'lambda:getfunctionurlconfig',
    'access_level': READ,
  },
  'lambda:listfunctionurlconfigs': {
    'action': 'lambda:listfunctionurlconfigs',
    'access_level': LIST,
  },
  'events:listendpoints': {
    'action': 'events:listendpoints',
    'access_level': LIST,
  },
  'events:describeendpoint': {
    'action': 'describeendpoint:describeendpoint',
    'access_level': READ,
  },
}



import re
import sys

from policy_sentry.querying.actions import get_actions_for_service
from policy_sentry.querying.actions import get_actions_matching_arn_type
from policy_sentry.querying.actions import get_action_data

from aws_iam_utils import checks
from aws_iam_utils.util import create_policy
from aws_iam_utils.util import statement
from aws_iam_utils.constants import READ, LIST, WRITE

def generate_read_only_policy_for_service(service_name, use_wildcard_verbs=True):
    """Generates an IAM policy that grants read-only access to all of the given service."""
    return generate_policy_for_service(service_name, [ LIST, READ ], use_wildcard_verbs=use_wildcard_verbs)

def generate_list_only_policy_for_service(service_name, use_wildcard_verbs=True):
    """Generates an IAM policy that grants list-only access to all of the given service."""
    return generate_policy_for_service(service_name, [ LIST ], use_wildcard_verbs=use_wildcard_verbs)

def generate_read_write_policy_for_service(service_name, use_wildcard_verbs=True):
    """Generates an IAM policy that grants read-write access to all of the given service."""
    return generate_policy_for_service(service_name, [ LIST, READ, WRITE ], use_wildcard_verbs=use_wildcard_verbs)

def generate_read_only_policy_for_service_arn_type(service_name, arn_type):
    """Generates an IAM policy that grants read-only access to all of the given service."""
    return generate_policy_for_service_arn_type(service_name, arn_type, [ LIST, READ ])

def generate_list_only_policy_for_service_arn_type(service_name, arn_type):
    """Generates an IAM policy that grants list-only access to all of the given service."""
    return generate_policy_for_service_arn_type(service_name, arn_type, [ LIST ])

def generate_read_write_policy_for_service_arn_type(service_name, arn_type):
    """Generates an IAM policy that grants read-write access to all of the given service."""
    return generate_policy_for_service_arn_type(service_name, arn_type, [ LIST, READ, WRITE ])

def generate_full_policy_for_service(*service_name):
    """Generates an IAM policy that grants full access to all of the given service."""
    return create_policy({
        "Effect": "Allow",
        "Action": [ f"{s}:*" for s in service_name ],
        "Resource": "*",
    })

def generate_policy_for_service_arn_type(service_name, arn_type, reqd_access_levels):
    """
    Generates an IAM policy that grants the given level of access to a specific ARN type within the given AWS service.

    Use `policy_sentry query arn-table --service <service>` to query ARN types available for a service.

    If use_wildcard_verbs is True (the default), the generator will try to use use verb wildcards (e.g. "s3:Get*") rather than full action names to keep the policy short and readable.

    The generator always verifies that the generated policy only grants permissions in the given access level. If this check fails when use_wildcard_verbs is True, the full action list is returned. If the check still fails with the full action list, an AssertionError is raised, and this indicates an underlying bug in the policy data driving aws-iam-utils.
    """
    service_actions = get_actions_matching_arn_type(service_name, arn_type)

    # use_wildcard_verbs is False here as it'll always fail (verb-based 
    # wildcards will always matching more than one ARN type)
    return __generate_and_validate_policy_from_actions(service_actions, service_name, reqd_access_levels, use_wildcard_verbs=False)

def generate_policy_for_service(service_name, reqd_access_levels, use_wildcard_verbs=True):
    """
    Generates an IAM policy that grants the given level of access to all of the given AWS service.

    If use_wildcard_verbs is True (the default), the generator will try to use use verb wildcards (e.g. "s3:Get*") rather than full action names to keep the policy short and readable.

    The generator always verifies that the generated policy only grants permissions in the given access level. If this check fails when use_wildcard_verbs is True, the full action list is returned. If the check still fails with the full action list, an AssertionError is raised, and this indicates an underlying bug in the policy data driving aws-iam-utils.
    """
    service_actions = get_actions_for_service(service_name)

    return __generate_and_validate_policy_from_actions(service_actions, service_name, reqd_access_levels, use_wildcard_verbs)


def __generate_and_validate_policy_from_actions(service_actions, service_name, reqd_access_levels, use_wildcard_verbs):
    matching_actions = []
    policy = None

    for action in service_actions:
        action_service, action_name = action.split(':')

        # iterate through each action and pull out read-only actions
        action_output = get_action_data(action_service, action_name)

        if action_output is False:
            raise ValueError(f'invalid action: {action_name}')

        for action_output_action in action_output[service_name]:
            if action_output_action["access_level"] in reqd_access_levels and action_output_action['action'] not in matching_actions:
                matching_actions.append(action_output_action["action"])

    if use_wildcard_verbs:
        # In this mode, we deduce the 'verb' (first word, assuming camel case) of each
        # action in the list and try to shorten the list to those verbs, wildcarded.
        # This approach makes for way shorter and easier-to-read policies.
        wildcarded_matching_actions = []

        for action in matching_actions:
            action_service, action_name = action.split(':')

            parts = list(filter(lambda x: len(x)>0, re.split('([A-Z])', action_name, 2)))

            if len(parts) >= 4:
                # 'DescribeJobAspectsPolicy' -> ['D', 'escribe', 'J', 'obAspectsPolicy']
                verb = parts[0]+parts[1]
                wildcarded_verb = f'{service_name}:{verb}*'

                if 'Policy' not in action_name and 'Tagging' not in action_name and wildcarded_verb not in wildcarded_matching_actions:
                    wildcarded_matching_actions.append(wildcarded_verb)

            else:
                wildcarded_matching_actions.append(action)

        policy = create_policy(statement(actions=wildcarded_matching_actions, resource='*'))

    else:
        policy = create_policy(statement(actions=matching_actions, resource='*'))

    if use_wildcard_verbs:
        # in this mode, check the policy is not too permissive and if so,
        # fall back to the full action list
        if not checks.policy_has_only_these_access_levels(policy, reqd_access_levels):
            policy = create_policy(statement(actions=matching_actions, resource='*'))

    assert checks.policy_has_only_these_access_levels(policy, reqd_access_levels)

    return policy



# Access levels as they appear in the raw AWS IAM data returned by get_action_data
# (see https://raw.githubusercontent.com/salesforce/policy_sentry/master/policy_sentry/shared/data/iam-definition.json)
READ = "Read"
LIST = "List"
WRITE = "Write"
TAGGING = "Tagging"
PERMISSIONS = "Permissions management"

ALL_ACCESS_LEVELS = [ READ, LIST, WRITE, TAGGING, PERMISSIONS ]

WILDCARD_ARN_TYPE = '*'



import json
from itertools import chain

from aws_iam_utils.util import extract_policy_permission_items
from aws_iam_utils.util import create_policy

def combine_policy_statements(*policies):
    """
    Merges the Statements for all the given policies into a single policy.
    """

    if len(policies) == 0:
        return create_policy()

    return create_policy(
        *list(chain(*[ p['Statement'] for p in policies ])),
        version=policies[0]["Version"],
    )

def collapse_policy_statements(*policies):
    """
    Attempts to merge policy statements together as far as possible, in order to simplify and shorten your policy. All statements with equal Effect, Condition, Principal and Resource keys will have their Actions merged together.
    """

    # create a single policy with all Statements combined together
    combined_policy = combine_policy_statements(*policies)

    items = extract_policy_permission_items(combined_policy)

    # to combine, we group all actions by their effect/resource/condition/principal,
    # and then generate a new policy with statements for each unique combination
    # of those
    actions_by_qualifiers = {}

    for item in items:
        # turn into json so we can use nested dicts/lists/etc as dict keys
        k = json.dumps([item['effect'], item['condition'], item['resource'], item['principal']])

        if not k in actions_by_qualifiers:
            actions_by_qualifiers[k] = []

        actions_by_qualifiers[k].append(item['action'])

    new_policy_statements = []
    for qualifiers, actions in actions_by_qualifiers.items():
        new_statement = {}
        qualifiers_loaded = json.loads(qualifiers)

        for k, v in {
            "Effect": qualifiers_loaded[0],
            "Condition": qualifiers_loaded[1],
            "Resource": qualifiers_loaded[2],
            "Principal": qualifiers_loaded[3],
        }.items():
            if v is not None: new_statement[k] = v

        # finally, remove duplicate actions while retaining original order
        # https://stackoverflow.com/a/25560184
        deduped_actions = sorted(set(actions), key=lambda x: actions.index(x))

        new_statement["Action"] = deduped_actions

        new_policy_statements.append(new_statement)

    return {
        "Version": combined_policy["Version"],
        "Statement": new_policy_statements
    }



import pathlib
from setuptools import setup
from setuptools import find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="aws-iam-utils",
    version="1.5.1",
    description="AWS IAM utility library",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jtyers/aws-iam-utils",
    author="Jonny Tyers",
    author_email="jonny@jonnytyers.co.uk",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["policy_sentry", "policyuniverse"],
    #entry_points={
    #},
)



def simplify_policy(p):
  """For the given policy, simplify any one-item arrays into straight strings, for Actions, Principals and Resources."""

  statements = p['Statement']

  for statement in statements:
    for k in [ "Action", "Resource" ]:
      if k in statement:
        if type(statement[k]) is list and len(statement[k]) == 1:
            statement[k] = statement[k][0]

    if 'Principal' in statement:
        principal = statement['Principal']

        for k in [ "AWS", "Service" ]:
          if k in principal:
            if type(principal[k]) is list and len(principal[k]) == 1:
                principal[k] = principal[k][0]

  return p



from policyuniverse.expander_minimizer import expand_policy
from policy_sentry.querying.actions import get_action_data
from aws_iam_utils.action_data_overrides import ACTION_DATA_OVERRIDES

def create_policy(*statements, version="2012-10-17"):
    """Shortcut function to create a policy with the given statements."""
    return {
        "Version": version,
        "Statement": list(statements)
    }

def statement(effect="Allow", actions=[], resource=None, condition=None, principal=None):
    """Shortcut function to create a Statement with the given properties."""
    st = {}

    for k, v in {
        "Effect": effect,
        "Action": actions,
        "Resource": resource,
        "Principal": principal,
        "Condition": condition,
    }.items():
        if v is not None: st[k] = v

    return st


def extract_policy_permission_items(policy, allow_unsupported=False):
    """
    For every individual permission granted, we build a list of
    { permission, resource, condition, principal } ("permission items").

    The policy is always expanded via expand_policy() first.

    This is useful for comparisons. Currently it does NOT support NotAction, NotPrincipal, NotResource keys. The presences of those keys will result in an exception, unless allow_unsupported is True.
    """

    items = []

    policy_expanded = policy

    for statement in policy_expanded["Statement"]:
        if not allow_unsupported:
            for k in [ "NotAction", "NotPrincipal", "NotResource" ]:
                if k in statement:
                    raise ValueError(f'Policy key {k} is not supported by extract_policy_permission_items() and will be ignored. To ignore this error, call extract_policy_permission_items with allow_unsupported=True.')


        for k in [ "Action", "Resource" ]:
            if type(statement.get(k)) is str:  # turn into list
                statement[k] = [statement[k]]
            
        effect = statement.get("Effect")
        condition = statement.get("Condition")
        principal = statement.get("Principal")

        for resource in statement.get("Resource", [None]):
            for action in statement["Action"]:
                items.append({
                    "effect": effect,
                    "action": action.lower(),
                    "resource": resource,
                    "condition": condition,
                    "principal": principal,
                })

    return items


def dedupe_list(lst):
    return sorted(set(lst), key=lambda x: lst.index(x))

def dedupe_policy(policy):
    """Deduplicates all Actions, Principals and Resources in the given policy."""
    for statement in policy["Statement"]:
        for k in [ "Action", "Resource" ]:
            if type(statement.get(k)) is list:
                statement[k] = dedupe_list(statement[k])
        
        principal = statement.get("Principal")
        for k in [ "AWS", "Service" ]:
            if type(principal.get(k)) is list:
                principal[k] = dedupe_list(principal[k])

    return policy

def get_action_data_with_overrides(service_name, action_name):
    full_action_name = f'{service_name}:{action_name.lower()}'
    if full_action_name in ACTION_DATA_OVERRIDES:
        return {
            service_name: [
                ACTION_DATA_OVERRIDES[full_action_name]
            ]
        }

    return get_action_data(service_name, action_name)
